package rabbit.html;

import java.io.UnsupportedEncodingException;

/** This is a class that is used to parse a block of HTML code into
 *  separate tokens. This parser uses a recursive descent approach.
 *
 * @author <a href="mailto:robo@khelekore.org">Robert Olofsson</a>
 */
public class HtmlParser {
    /** The actual data to parse. */
    private String pagepart = null;    
    /** The size of the data to parse. */
    private int dlength;
    /** The char set */
    String cs;
    /** The type of the next token. */
    private int nextToken = START;
    /** Index of the parse. */
    private int scanIndex = 0;
    /** The current tag started here. */
    private int tagStart = 0;
    /** The current value as a String. */
    private String stringValue = null;
    /** the current start of string. */
    private int stringLength = -1;

    /** True if were in a Tag, false otherwise. */
    private boolean tagmode = false;
    /** The last tag started here. */
    private int lastTagStart = 0;
    /** The block we have. */
    private HtmlBlock block;
    /** A pending comment (script or style data). */
    private Token pendingComment = null;
    
    /** This indicates the start of a block. */
    public final static int START        = 0;
    /** This indicate a String value was found. */
    public final static int STRING       = 1;
    /** This is a Single Quoted String a 'string' */
    public final static int SQSTRING     = 2;
    /** This is a Double Quoted String a &quot;string&quot; */
    public final static int DQSTRING     = 3;
    /** This is the character ''' */
    public final static int SINGELQUOTE  = 4;
    /** This is the character '"' */
    public final static int DOUBLEQUOTE  = 5;
    /** Less Than '<' */
    public final static int LT           = 6;
    /** More Than '>' */
    public final static int MT           = 7;
    /** Equals '=' */
    public final static int EQUALS       = 8;
    /** A HTML comment &quot;&lt;&#33-- some text --&gt;&quot; */
    public final static int COMMENT      = 9;
    /** A HTML script */
    public final static int SCRIPT       = 10;

    /** This indicates the end of a block. */
    public final static int END          = 100;
    /** Unknown token. */
    public final static int UNKNOWN      = 1000;

    /** Create a new HTMLParser */
    public HtmlParser (String cs) {
	this.cs = cs;
    }

    /** Set the data block to parse.
     * @param page the block to parse.
     */
    public void setText (byte[] page) {
	setText (page, 0, page.length);
    }

    /** Set the data block to parse.
     * @param page the block to parse.
     * @param length the length of the data.
     */
    public void setText (byte[] page, int startIndex, int length) {
	// convert all to char
	try {
	    this.pagepart = new String (page, cs);
	} catch (UnsupportedEncodingException e) {
	    return;
	}
	scanIndex = startIndex;
	this.dlength = pagepart.length ();
    }
    
    /** Get a String describing the token.
     * @param token the token type (like STRING).
     * @return a String describing the token (like &quot;STRING&quot;)
     */
    private String getTokenString (int token) {
	switch (token) {
	case START:
	    return "START";
	case STRING:
	    return "STRING";
	case SQSTRING:
	    return "SQSTRING";
	case DQSTRING:
	    return "DQSTRING";
	case SINGELQUOTE:
	    return "SINGELQUOTE";
	case DOUBLEQUOTE:
	    return "DOUBLEQUOTE";
	case LT:
	    return "LT";
	case MT:
	    return "MT";
	case EQUALS:
	    return "EQUALS"; 
	case COMMENT:
	    return "COMMENT";
	case END:
	    return "END";
	case UNKNOWN:
	    return "UNKNOWN";
	default:
	    return "unknown";
	}
    }
    
    /** Scan a String from the block.
     * @throws HtmlParseException if an error occurs.
     * @return STRING
     */
    private int scanString () { 
	int endindex = dlength;
	int startindex = scanIndex - 1;
	
	if (tagmode) {
	loop:
	    while (scanIndex < dlength) {
		switch (pagepart.charAt (scanIndex)) {
		case ' ':
		case '\t':
		case '\n':
		case '\r':
		case '\"':
		case '\'':
		case '<':
		case '>':
		case '=':
		    endindex = scanIndex;
		    break loop;
		default:
		    scanIndex++;
		}
	    }
	} else {
	    while (scanIndex < dlength) {
		if (pagepart.charAt (scanIndex) == '<') {
		    endindex = scanIndex;
		    break;
		} else
		    scanIndex++;
	    }
	}
	if (tagmode) {
	    stringValue = pagepart.substring (startindex, endindex);
	} else {
	    stringLength = (endindex - startindex);
	}	    
	return STRING;
    }
    
    /** Scan a quoted tring from the block. The first character is
     *  treated as the quotation character.
     * @throws HtmlParseException if an error occurs.
     * @return SQSTRING, DQSTRING or UNKNOWN (for strange quotes).
     */
    private int scanQuotedString () {
	int endindex = -1;
	int startindex = scanIndex - 1;
	char start = pagepart.charAt (startindex);
	while (scanIndex < dlength) {
	    if (pagepart.charAt (scanIndex++) == start) {
		endindex = scanIndex;
		break;
	    }
	}
	if (endindex == -1) {
	    block.setRest (lastTagStart);
	    return END;
	}
	int e = (endindex < dlength ? endindex : dlength) ;
	stringValue = pagepart.substring (startindex, e);
	switch (start) {
	case (int)'\'':
	    return SQSTRING;
	case (int)'"':
	    return DQSTRING;
	default:
	    return UNKNOWN;
	}
    }

    /** Is this tag a comment?
     * @return true if the block(at current index) starts with !--,
     *  false otherwise.
     */
    private boolean isComment () {
	if (scanIndex + 3 >= dlength)
	    return false;
	return pagepart.startsWith( "!--", scanIndex);
    }
    
    /** Scan a comment from the block, that is the string up to and
     *	including &quot;-->&quot;.
     * @return COMMENT or END.
     */
    private int scanComment () throws HtmlParseException {
	int startvalue = scanIndex - 1;
	int i = -1;
	int j = scanIndex;
	while (j + 2 < dlength) {
	    if (pagepart.startsWith ("-->", j)) {
		i = j;
		break;
	    }
	    j++;
	}
	if (i > -1) {
	    scanIndex = i + 2;
	    nextToken = MT;
	    match (MT);
	    stringLength = scanIndex - startvalue;
	    return COMMENT;
	}
	block.setRest (startvalue);
	return END;
    }

    /** Match the token with next token and scan the (new)next token.
     * @param token the token to match.
     * @return the next token.
     */
    private int match (int token) throws HtmlParseException {
	int ts;
	
	if (nextToken != token)
	    throw new HtmlParseException ("Token: " + getTokenString (token) +
					  " != " + getTokenString (nextToken));

	if (pendingComment != null) {
	    nextToken = LT;
	    pendingComment = null;
	    return SCRIPT;
	}
	while (scanIndex < dlength) {
	    tagStart = scanIndex;
	    stringValue = null;
	    switch (pagepart.charAt (scanIndex++)) {
	    case ' ':
	    case '\t':
	    case '\n':
	    case '\r':
		// TODO evaluate strategy here, a continue here may result 
		// in RabbIT cutting out whitespaces from the html-page.
		continue;
	    case '<':
		ts = tagStart;
		if (isComment ()) {
		    nextToken = scanComment ();
		    tagStart = ts;
		    return nextToken;
		} else 
		    return nextToken = LT;
	    case '>':
		return nextToken = MT;
	    case '=':
		stringValue = "=";
		return nextToken = EQUALS;
	    case '"':
		if (tagmode)
		    return nextToken = scanQuotedString ();	
		// else fallthrough...
	    case '\'':
		if (tagmode)
		    return nextToken = scanQuotedString ();
		// else fallthrough...
	    default:
		return nextToken = scanString ();
	    }
	}

	return nextToken = END;
    }

    /** Scan a value from the block.
     * @return the value or null.
     */
    private String value () throws HtmlParseException {
	while (nextToken == EQUALS) {
	    match (EQUALS);
	    if (nextToken == STRING ||
		nextToken == SQSTRING || 
		nextToken == DQSTRING) {
		String val = stringValue;
		match (nextToken);
		return val;		   
	    }	    
	    return "";
	}
	return null;
    }

    private void setPendingComment (Token comment) {
	nextToken = SCRIPT;
	this.pendingComment = comment;
	tagStart = pendingComment.getStartIndex ();
	stringLength = pendingComment.getLength ();
    }

    /** Scan an argument list from the block.
     * @param tag the Tag that have the arguments.
     */
    private void arglist (Tag tag) throws HtmlParseException {
	String key = null;
	
	//System.err.println ("parsing arglist for tag: '" + tag + "'");
	while (true) {
	    //System.err.println ("nextToken: " + nextToken + " => " + 
	    //                    getTokenString (nextToken));
	    switch (nextToken) {
	    case MT:
		tagmode = false;
		// ok, this is kinda ugly but safer this way
		if (tag.getLowerCaseType () != null && 
		    (tag.getLowerCaseType ().equals ("script") ||
		     tag.getLowerCaseType ().equals ("style"))) {
		    Token text = scanCommentUntilEnd (tag.getLowerCaseType ());
		    if (text != null) {
			setPendingComment (text);
		    } else {
			tagmode = false;	    
			return;
		    }
		} else {		
		    match (MT);
		}
		return;
	    case STRING:
		key = stringValue;
		match (STRING);
		String value = value ();
		tag.addArg (key, value, false);
		break;
	    case END:
		return;
	    case DQSTRING:
		String ttype = tag.getType ();
		if (ttype != null && ttype.charAt (0) == '!') {
		    // Handle <!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 
		    //         Transitional//EN">
		    // and similar cases.
		    tag.addArg (stringValue, null, false);
		    match (nextToken);
		} else {
		    //System.err.println ("hmmm, strange arglist..: " + tag);
		    // this is probably due to something like: 
		    // ... framespacing="0"">
		    // we backstep and change the second '"' to a blank and 
		    // restart from that point...
		    scanIndex -= stringValue.length ();
		    pagepart= pagepart.substring(0, scanIndex) + ' ' + pagepart.substring (scanIndex + 1);
		    match (nextToken);
		    tag.getToken ().setChanged (true);
		}
		break;
	    case LT:   
		// I consider this an error (in the html of the page) 
		// but handle it anyway.
		String type = tag.getLowerCaseType ();	       
		if (type == null ||              // <<</font.... etc
		    (type.equals ("table") &&    // <table.. width=100% <tr>
		     stringValue == null)) {
		    tagmode = false;
		    return;
		}
		// fall through.
	    default:
		//System.err.println ("hmmm, default arglist..: " + tag);
		// this is probably due to something like: 
		// <img src=someimagead;ad=40;valu=560>
		// we will break at '=' and split the tag to something like: 
		// <img src=someimagead;ad = 40;valu=560> if we change it.
		// the html is already broken so should we fix it? 
		// we ignore for now..
		if (stringValue != null)
		    tag.addArg (stringValue, null, false);
		match (nextToken);
	    }
	}
    }
    /** Is this tag at the scan position? */
    private boolean sameTag (String tag, int j) {
	String str;
	try {
	    str = pagepart.substring(j, j + tag.length ());
	} catch (IndexOutOfBoundsException e) {
	    return false;
	}
	return (0 == tag.compareToIgnoreCase (str));
    }

    private Token scanCommentUntilEnd (String tag) {
	int len = tag.length ();
	int startvalue = scanIndex;
	int i = -1;
	int j = scanIndex;
	while (j + 1 + len < dlength) {
	    if (pagepart.charAt (j)     == '<' &&
		pagepart.charAt (j + 1) == '/' &&
		sameTag (tag, j + 2)) {
		i = j;
		break;
	    }
	    j++;
	}
	if (i > -1) {
	    stringLength = j - startvalue;
	    scanIndex = j;
	    Token text = 
		new Token (pagepart.substring(startvalue, startvalue+stringLength), TokenType.COMMENT, startvalue); 
	    return text;
	}
	block.setRest (lastTagStart);
	return null;
    }

    /** Scan a tag from the block.
     * @param ltagStart the index of the last tag started.
     */
    private void tag (int ltagStart) throws HtmlParseException {		
	Tag tag = new Tag ();
	Token token = new Token (tag, false);
	switch (nextToken) {
	case STRING: 
	    tag.setType (stringValue);
	    match (STRING);
	    arglist (tag);
	    if (tagmode) {
		block.setRest (lastTagStart);
	    } else {
		if (block.restSize () == 0) {
		    token.setStartIndex (ltagStart);
		    block.addToken (token);
		}
	    }
	    break;
	case MT:
	    tagmode = false;			
	    match (MT);
	    break;
	case END:
	    block.setRest (lastTagStart);
	    tagmode = false;	    
	    return;
	default: 
	    arglist (tag);
	}
    }

    /** Scan a page from the block.
     */
    private void page () throws HtmlParseException {
	while (block.restSize () == 0) {
	    switch (nextToken) {
	    case END:
		return;
	    case LT:
		lastTagStart = tagStart;
		tagmode = true;
		match (LT);
		tag (lastTagStart);
		break;
	    case COMMENT: 
		block.addToken (new Token (pagepart.substring(tagStart, tagStart+stringLength), TokenType.COMMENT, tagStart)); //zzz -1?
		match (COMMENT);
		break;
	    case SCRIPT:
		block.addToken (new Token (pagepart.substring(tagStart, tagStart+stringLength), TokenType.SCRIPT, tagStart)); //zzz -1?
		match (SCRIPT);
		break;
	    case STRING:
	    default:
		block.addToken (new Token (pagepart.substring(tagStart, tagStart+stringLength), TokenType.TEXT, tagStart)); //zzz -1?
		match (nextToken);
	    }
	}
    }


    /** Get a HtmlBlock from the pagepart given.
     */
    public HtmlBlock parse () throws HtmlParseException {
	block = new HtmlBlock (pagepart, dlength, cs);
	nextToken = START;
	match (START);
	page ();
	
	return block;
    }
}
